#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/utsname.h>

#include "haps.h"
#include "md5.h"
#include "list.h"
#include "module.h"

struct isapnpDevice
{
  /* common fields */
  char *device;			/* device file associated with it */
  char *driver;			/* driver to load, if any */
  char *desc;			/* a description */
  int detached;

  char *deviceId;
  char *pdeviceId;
  char *compat;
  int native;
  int active;
  int cardnum;
  int logdev;
  int *io;
  int *irq;
  int *dma;
  int *mem;
};


static struct isapnpDevice *isapnpDeviceList = NULL;
static int numIsapnpDevices = 0;

static char *soundlist[] = {
  "sb",
  "cs4232",
  "ad1816",
  "ad1848",
  "sscape",
  "gus",
  "opl3sa2",
  "mad16",
  "awe_wave",
  NULL
};

static char *netlist[] = {
  "3c509",
  "3c515",
  "ne",
  "sb1000",
  "smc-ultra",
  NULL
};

static char *scsilist[] = {
  "aha1542",
  "g_NCR5380",
  NULL
};

static char *
demangle (int vendor, int device)
{
  static char ret[8];

  sprintf (ret, "%c%c%c%x%x%x%x",
	   'A' + ((vendor >> 2) & 0x3f) - 1,
	   'A' + (((vendor & 3) << 3) | ((vendor >> 13) & 7)) - 1,
	   'A' + ((vendor >> 8) & 0x1f) - 1,
	   (device >> 4) & 0x0f,
	   device & 0x0f, (device >> 12) & 0x0f, (device >> 8) & 0x0f);
  return ret;
}

static int
devCmp (const void *a, const void *b)
{
  const struct isapnpDevice *one = a;
  const struct isapnpDevice *two = b;
  int x = 0, y = 0;

  x = strcmp (one->deviceId, two->deviceId);
  if (one->pdeviceId && two->pdeviceId)
    y = strcmp (one->pdeviceId, two->pdeviceId);
  else
    y = one->pdeviceId - two->pdeviceId;
  if (x)
    return x;
  else if (y)
    return y;
  return 0;
}

int
isapnpReadDrivers (char *filename)
{
  int fd;
  char path[256];
  struct utsname utsbuf;
  int id1, id2, id3, id4;
  struct isapnpDevice key, *nextDevice;
  char *ident, *pident;
  char *buf, *start, *next, *ptr, *module;

  uname (&utsbuf);
  snprintf (path, 255, "/lib/modules/%s/modules.isapnpmap", utsbuf.release);
  fd = open (path, O_RDONLY);
  if (fd < 0)
  {
    fd = open ("/etc/modules.isapnpmap", O_RDONLY);
    if (fd < 0)
    {
      fd = open ("/modules/modules.isapnpmap", O_RDONLY);
      if (fd < 0)
      {
	fd = open ("./modules.isapnpmap", O_RDONLY);
	if (fd < 0)
	  return -1;
      }
    }
  }

  start = buf = bufFromFd (fd);

  nextDevice = isapnpDeviceList + numIsapnpDevices;

  while (*buf)
  {
    next = buf;
    while (*next && *next != '\n')
      next++;
    if (*next)
    {
      *next = '\0';
      next++;
    }
    ptr = buf;
    if (*ptr == '#')
    {
      buf = next;
      continue;
    }

    while (*ptr && !isspace (*ptr))
      ptr++;
    if (*ptr)
    {
      *ptr = '\0';
      ptr++;
    }
    while (isspace (*ptr))
      ptr++;
    module = strdup (buf);
    buf = ptr;

    while (*ptr && !isspace (*ptr))
      ptr++;
    if (*ptr)
    {
      *ptr = '\0';
      ptr++;
    }
    while (isspace (*ptr))
      ptr++;
    id1 = strtoul (buf, NULL, 16);
    buf = ptr;

    while (*ptr && !isspace (*ptr))
      ptr++;
    if (*ptr)
    {
      *ptr = '\0';
      ptr++;
    }
    while (isspace (*ptr))
      ptr++;
    id2 = strtoul (buf, NULL, 16);
    buf = ptr;

    while (*ptr && !isspace (*ptr))
      ptr++;
    if (*ptr)
    {
      *ptr = '\0';
      ptr++;
    }
    while (isspace (*ptr))
      ptr++;
    buf = ptr;

    while (*ptr && !isspace (*ptr))
      ptr++;
    if (*ptr)
    {
      *ptr = '\0';
      ptr++;
    }
    while (isspace (*ptr))
      ptr++;
    id3 = strtoul (buf, NULL, 16);
    buf = ptr;

    while (*ptr && !isspace (*ptr))
      ptr++;
    if (*ptr)
    {
      *ptr = '\0';
      ptr++;
    }
    while (isspace (*ptr))
      ptr++;
    id4 = strtoul (buf, NULL, 16);

    pident = strdup (demangle (id1, id2));
    ident = strdup (demangle (id3, id4));
    key.deviceId = ident;
    key.pdeviceId = pident;

    nextDevice = bsearch (&key, isapnpDeviceList, numIsapnpDevices,
			  sizeof (struct isapnpDevice), devCmp);

    if (!nextDevice)
    {
      isapnpDeviceList = realloc (isapnpDeviceList,
				  (numIsapnpDevices + 1) *
				  sizeof (struct isapnpDevice));
      nextDevice = isapnpDeviceList + numIsapnpDevices;
      memset (nextDevice, '\0', sizeof (struct isapnpDevice));
      nextDevice->driver = module;
      nextDevice->deviceId = ident;
      nextDevice->pdeviceId = pident;
      nextDevice++;
      numIsapnpDevices++;
      qsort (isapnpDeviceList, numIsapnpDevices,
	     sizeof (*isapnpDeviceList), devCmp);
    }
    else
    {
      free (ident);
      free (pident);
      free (module);
    }
    buf = next;
  }
  free (start);
  return 0;
}

void
isapnpFreeDrivers (void)
{
  int x;

  if (isapnpDeviceList)
  {
    for (x = 0; x < numIsapnpDevices; x++)
    {
      if (isapnpDeviceList[x].deviceId)
	free (isapnpDeviceList[x].deviceId);
      if (isapnpDeviceList[x].driver)
	free (isapnpDeviceList[x].driver);
    }
    free (isapnpDeviceList);
    isapnpDeviceList = NULL;
    numIsapnpDevices = 0;
  }
}

static void
setDriverAndClass (struct node *dev)
{
  struct isapnpDevice key, *searchdev;
  int x;

  key.deviceId = nodegetprop (dev, "DEVICEID");
  key.pdeviceId = nodegetprop (dev, "PDEVICEID");
  searchdev = bsearch (&key, isapnpDeviceList, numIsapnpDevices,
		       sizeof (struct isapnpDevice), devCmp);
  if (!searchdev)
  {
    key.pdeviceId = demangle (0xffff, 0xffff);
    searchdev = bsearch (&key, isapnpDeviceList, numIsapnpDevices,
			 sizeof (struct isapnpDevice), devCmp);
  }
  if (!searchdev && nodegetprop (dev, "COMPAT"))
  {
    key.deviceId = nodegetprop (dev, "COMPAT");
    searchdev = bsearch (&key, isapnpDeviceList, numIsapnpDevices,
			 sizeof (struct isapnpDevice), devCmp);
  }
  if (searchdev)
  {
    nodesetprop (dev, "DRIVER", searchdev->driver);
    nodesetprop (dev, "NATIVE", "1");
  }
  else
  {
    nodesetprop (dev, "NATIVE", "0");
    if (!strstr (nodegetprop (dev, "DESC"), "IDE")
	&& !strstr (nodegetprop (dev, "DESC"), "ATAPI")
	&& !strstr (nodegetprop (dev, "DESC"), "CD-ROM")
	&& !strstr (nodegetprop (dev, "DESC"), "CDROM"))
    {
      if (!strncmp (nodegetprop (dev, "PDEVICEID"), "CSC", 3)
	  && strcmp (nodegetprop (dev, "PDEVICEID"), "CSC6040"))
	nodesetprop (dev, "DRIVER", "cs4232");
      if (!strncmp (nodegetprop (dev, "PDEVICEID"), "ENS", 3))
      {
	if (!strstr (nodegetprop (dev, "DESC"), "VIVO"))
	  nodesetprop (dev, "DRIVER", "ad1848");
	else
	  nodesetprop (dev, "DRIVER", "sscape");
      }
      if (!strncmp (nodegetprop (dev, "PDEVICEID"), "GRV", 3))
	nodesetprop (dev, "DRIVER", "gus");
      if (!strncmp (nodegetprop (dev, "PDEVICEID"), "AZT", 3))
	nodesetprop (dev, "DRIVER", "ad1848");
      if (!strncmp (nodegetprop (dev, "PDEVICEID"), "OPT", 3))
	nodesetprop (dev, "DRIVER", "mad16");
      if (!strncmp (nodegetprop (dev, "PDEVICEID"), "CMI", 3))
	nodesetprop (dev, "DRIVER", "ad1848");
    }
  }
  for (x = 0; soundlist[x]; x++)
  {
    if (!strcmp (soundlist[x], nodegetprop (dev, "DRIVER")))
      nodesetprop (dev, "CLASS", "AUDIO");
  }
  for (x = 0; netlist[x]; x++)
  {
    if (!strcmp (netlist[x], nodegetprop (dev, "DRIVER")))
    {
      nodesetprop (dev, "CLASS", "NETWORK");
      nodesetprop (dev, "DEVICE", "eth");
    }

  }
  for (x = 0; scsilist[x]; x++)
  {
    if (!strcmp (scsilist[x], nodegetprop (dev, "DRIVER")))
      nodesetprop (dev, "CLASS", "SCSI");
  }
}

int
isapnpReadResources (char *line, int base)
{
  int *ret = NULL, x = 0;
  char *ptr;

  do
  {
    ptr = strstr (line, ",");
    if (ptr)
      *ptr = '\0';
    x++;
    ret = realloc (ret, (x + 2) * sizeof (int));
    ret[x - 1] = strtoul (line, NULL, base);
    ret[x] = -1;
    if (ptr)
      line = ptr + 1;
  }
  while (ptr);
  return *ret;
}

struct node *
probe (struct node *devices, int params)
{
  char *pnpbuf = NULL;
  char *start, *current, *ptr;
  struct node *dev = NULL;
  char pdev[10], pdesc[64];
  char *pnpdesc;
  int cardnum;
  char buf_isa[4096], buf[0x100];
  int fd, bytes = 0;
  int init_list = 0;

  devices=nodecondremove(devices,"BUS","ISAPNP",1);

  if (!isapnpDeviceList)
  {
    isapnpReadDrivers (NULL);
    init_list = 1;
  }
  fd = open ("./isapnp", O_RDONLY);
  if (fd == -1)
  {
    fd = open ("/proc/isapnp", O_RDONLY);
    if (fd == -1)
    {
      return devices;
    }
  }

  memset (buf_isa, '\0', 4096);
  while (read (fd, buf_isa, 4096) > 0)
  {
    pnpbuf = realloc (pnpbuf, bytes + 4096);
    memcpy (pnpbuf + bytes, buf_isa, 4096);
    bytes += 4096;
    memset (buf_isa, '\0', 4096);
  }
  close (fd);

  start = pnpbuf;

  while (start && *start)
  {
    current = start;
    while (*current && *current != '\n')
      current++;
    if (*current)
    {
      *current = '\0';
      current++;
    }
    if (!strncmp ("Card ", start, 5))
    {
      cardnum = atoi (start + 5) - 1;
      start += 8;
      if ((ptr = strstr (start, ":")))
      {
	*ptr = '\0';
	strncpy (pdev, start, 10);
	start = ptr + 1;
	if ((ptr = strstr (start, "'")))
	{
	  *ptr = '\0';
	  strncpy (pdesc, start, 64);
	  start = ptr + 1;
	}
      }
    }
    else if (!strncmp ("  Logical device", start, 16))
    {
      if (dev)
      {
	setDriverAndClass (dev);
        devices=nodeadd(devices,dev);
      }
      dev=nodealloc();
      nodesetprop(dev,"BUS","ISAPNP");
      sprintf(buf,"isapnp.%d",cardnum);
      nodesetprop(dev,"CONN",buf);
      nodesetpropint (dev, "CARDNUM", cardnum);
      nodesetprop (dev, "PDEVICEID", pdev);
      nodesetprop (dev, "DRIVER", "unknown");
      nodesetpropint (dev, "LOGDEV", atoi (start + 17));
      start += 20;
      if ((ptr = strstr (start, ":")))
      {
	*ptr = '\0';
	nodesetprop (dev, "DEVICEID", start);
	start = ptr + 1;
	if ((ptr = strstr (start, "'")))
	{
	  *ptr = '\0';
	  pnpdesc = strdup (start);
	}
      }
      sprintf (buf, "%s:%s", pdesc, pnpdesc);
      nodesetprop (dev, "DESC", buf);
      if (pnpdesc)
	free (pnpdesc);
    }
    else if (!strncmp ("    Device is active", start, 20))
    {
      nodesetprop (dev, "ACTIVE", "1");
    }
    else if (!strncmp ("    Compatible device", start, 21))
    {
      nodesetprop (dev, "COMPAT", start + 22);
    }
    else if (!strncmp ("    Active port", start, 15))
    {
      nodesetpropint (dev, "IO", isapnpReadResources (start + 16, 16));
    }
    else if (!strncmp ("    Active IRQ", start, 14))
    {
      nodesetpropint (dev, "IRQ", isapnpReadResources (start + 15, 10));	// <-CHECK ME!!!
    }
    else if (!strncmp ("    Active DMA", start, 14))
    {
      nodesetpropint (dev, "DMA", isapnpReadResources (start + 15, 10));	// <-CHECK ME!!! 
    }
    else if (!strncmp ("    Active memory", start, 17))
    {
      nodesetpropint (dev, "MEM", isapnpReadResources (start + 18, 16));
    }
    start = current;
  }
  if (dev)
  {
    setDriverAndClass (dev);
    devices=nodeadd(devices,dev);
  }
  free (pnpbuf);

  if (isapnpDeviceList && init_list)
    isapnpFreeDrivers ();

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
