#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "md5.h"
#include "list.h"
#include "module.h"

static char *procDirs[] = { "/proc/sys/dev/parport", "/proc/parport"
};

static struct node *
readProbeInfo (char *ppath)
{
  int lpfile, bytes;
  char *probebuf, *ptr;
  char *mfr, *model, *desc, *func, *modes;
  struct node *pardev;

  lpfile = open (ppath, O_RDONLY);
  if (lpfile == -1)
    return NULL;
  probebuf = calloc (8192, sizeof (char));
  bytes = read (lpfile, probebuf, 8192);
  if (bytes <= 0)
  {
    close (lpfile);
    return NULL;
  }

  mfr = model = desc = func = modes = NULL;
  ptr = probebuf;
  while (probebuf[0] != '\0')
  {
    while (*ptr && *ptr != '\n')
      ptr++;
    if (*ptr)
    {
      *ptr = '\0';
      /* strip trailing semicolon, too */
      *(ptr - 1) = '\0';
      ptr++;
    }
    if (!strncmp (probebuf, "MFG:", 4)
	|| !strncmp (probebuf, "MANUFACTURER:", 13))
    {
      mfr = strdup (strstr (probebuf, ":") + 1);
    }
    if (!strncmp (probebuf, "MDL:", 4) || !strncmp (probebuf, "MODEL:", 6))
    {
      model = strdup (strstr (probebuf, ":") + 1);
    }
    if (!strncmp (probebuf, "CLS:", 4) || !strncmp (probebuf, "CLASS:", 6))
    {
      func = strdup (strstr (probebuf, ":") + 1);
    }
    if (!strncmp (probebuf, "CMD:", 4)
	|| !strncmp (probebuf, "COMMAND SET:", 12))
    {
      modes = strdup (strstr (probebuf, ":") + 1);
    }
    if (!strncmp (probebuf, "DES:", 4)
	|| !strncmp (probebuf, "DESCRIPTION:", 12))
    {
      desc = strdup (strstr (probebuf, ":") + 1);
    }
    probebuf = ptr;
  }
  if (!strcmp (mfr, "Unknown vendor") && !strcmp (model, "Unknown device"))
    return NULL;

  pardev = nodealloc ();
  nodesetprop (pardev, "BUS", "PARALLEL");

  if (!desc)
  {
    desc = malloc (strlen (mfr) + strlen (model) + 2);
    if (mfr && model)
    {
      snprintf (desc, strlen (mfr) + strlen (model) + 2, "%s %s", mfr, model);
    }
  }
  nodesetprop (pardev, "DESC", desc);
  nodesetprop (pardev, "DRIVER", "unknown");
  nodesetprop (pardev, "PNPMFR", mfr);
  nodesetprop (pardev, "PNPMODEL", model);
  if (modes)
    nodesetprop (pardev, "PNPMODES", modes);
  if (desc)
    nodesetprop (pardev, "PNPDESC", desc);
  nodesetprop (pardev, "CLASS", "OTHER");
  if (func)
  {
    if (!strcmp (func, "PRINTER"))
    {
      nodesetprop (pardev, "CLASS", "PRINTER");
    }
    else if (!strcmp (func, "MODEM"))
    {
      nodesetprop (pardev, "CLASS", "MODEM");
    }
    else if (!strcmp (func, "NET"))
    {
      nodesetprop (pardev, "CLASS", "NET");
    }
    else if (!strcmp (func, "HDC"))
    {
      nodesetprop (pardev, "CLASS", "HD");
    }
    else if (!strcmp (func, "FDC"))
    {
      nodesetprop (pardev, "CLASS", "FLOPPY");
    }
    else if (!strcmp (func, "SCANNER"))
    {
      nodesetprop (pardev, "CLASS", "SCANNER");
    }
  };

  if (mfr)
    free (mfr);
  if (model)
    free (model);
  if (func)
    free (func);
  if (modes)
    free (modes);
  if (desc)
    free (desc);
  return pardev;
}

struct node *
probe (struct node *devices, int params)
{
  DIR *dir;
  struct dirent *dent;
  char path[256];
  int procdir = 0;
  struct node *pardev;
  char buf[0x100];
  MD5_CTX context;
  unsigned char digest[16];

  devices = nodecondremove (devices, "BUS", "PARALLEL", 1);

  dir = opendir (procDirs[0]);
  if (!dir)
  {
    dir = opendir (procDirs[1]);
    if (dir)
      procdir = 1;
  }
  else
  {
    procdir = 0;
  }
  if (dir)
  {
    while ((dent = readdir (dir)))
    {
      if (dent->d_name[0] == '.')
	continue;
      snprintf (path, 256, "%s/%s/autoprobe", procDirs[procdir],
		dent->d_name);
      pardev = readProbeInfo (path);
      if (pardev)
      {
	if (!strncmp (dent->d_name, "parport", 7))
	{
	  snprintf (path, 256, "/dev/lp%s", dent->d_name + 7);
	  sprintf (buf, "lp.%s", dent->d_name + 7);
	  nodesetprop (pardev, "CONN", buf);
	}
	else
	{
	  snprintf (path, 256, "/dev/lp%s", dent->d_name);
	  sprintf (buf, "lp.%s", dent->d_name);
	  nodesetprop (pardev, "CONN", buf);
	}


	nodesetprop (pardev, "DEVICE", path);

	MD5Init (&context);
	MD5UpdateS (&context, nodegetprop (pardev, "DESC"));
	MD5Final (digest, &context);

	sprintf (buf,
		 "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
		 digest[0], digest[1], digest[2], digest[3], digest[4],
		 digest[5], digest[6], digest[7], digest[8], digest[9],
		 digest[10], digest[11], digest[12], digest[13],
		 digest[14], digest[15]);

	nodesetprop (pardev, "MD5_ID", buf);
	devices = nodeadd (devices, pardev);
      }
    }
    closedir (dir);
  }
  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
